// Inspired by the DuckDuckGo Privacy Extension (https://github.com/duckduckgo/duckduckgo-privacy-extension)
// Original source for this file can be found here:
// https://github.com/duckduckgo/duckduckgo-privacy-extension/tree/develop/shared/js/background

/*
 * Copyright (C) 2012, 2016 DuckDuckGo, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* For ESLint: List any global identifiers used in this file below */
/* global require log module chrome */

'use strict';

/**
 * NOTE: this needs to be the first listener that's added
 *
 * on FF, we might actually miss the onInstalled event
 * if we do too much before adding it
 */
const utils = require('./utils.es6');


/**
 * REQUESTS
 */

const redirect = require('./redirect.es6');
const tabManager = require('./tab-manager.es6').tabManager;
const TabManager = require('./tab-manager.es6').TabManager;
const https = require('./https.es6');
const constants = require('./constants');

const requestListenerTypes = utils.getUpdatedRequestListenerTypes();

// Shallow copy of request types
// And add beacon type based on browser, so we can block it
chrome.webRequest.onBeforeRequest.addListener(
  redirect.handleRequest,
  {
    urls: ['<all_urls>'],
    types: requestListenerTypes,
  },
  ['blocking'],
);

chrome.webRequest.onHeadersReceived.addListener(
  (request) => {
    if (request.type === 'main_frame') {
      TabManager.updateTabUrl(request);
    }
  },
  {
    urls: ['<all_urls>'],
  },
);

// keep track of URLs that the browser navigates to.
//
// this is currently meant to supplement tabManager.updateTabUrl() above:
// tabManager.updateTabUrl only fires when a tab has finished loading with a 200,
// which misses a couple of edge cases like browser special pages
// and Gmail's weird redirect which returns a 200 via a service worker
chrome.webNavigation.onCommitted.addListener(
  (details) => {
    // ignore navigation on iframes
    if (details.frameId !== 0) {
      return;
    }

    const tab = tabManager.get({ tabId: details.tabId });

    if (!tab) {
      return;
    }

    tab.updateSite(details.url);
  },
);

/**
 * TABS
 */
//
chrome.tabs.onUpdated.addListener((id, info) => {
  TabManager.createOrUpdateTab(id, info);
});

chrome.tabs.onRemoved.addListener((id) => {
  // remove the tab object
  tabManager.delete(id);
});

/**
 * MESSAGES
 */

const DDGsettings = require('./settings.es6');
const browserWrapper = require('./chrome-wrapper.es6');

/**
 * ALARMS
 */

const httpsStorage = require('./storage/https.es6');
// recheck tracker and https lists every 12 hrs
chrome.alarms.create('updateHTTPSLists', { periodInMinutes: 12 * 60 });
// tracker lists / whitelists are 30 minutes
chrome.alarms.create('updateLists', { periodInMinutes: 30 });

chrome.alarms.onAlarm.addListener((alarmEvent) => {
  if (alarmEvent.name === 'updateHTTPSLists') {
    DDGsettings.ready().then(() => {
      httpsStorage.getLists(constants.httpsLists)
        .then(lists => https.setLists(lists))
        /* eslint-disable-next-line no-console */
        .catch(e => console.log(e));
    });
  }
});

/**
 * on start up
 */
const onStartup = () => {
  chrome.tabs.query({ currentWindow: true, status: 'complete' }, (savedTabs) => {
    for (let i = 0; i < savedTabs.length; i += 1) {
      const tab = savedTabs[i];
      if (tab.url) {
        tabManager.create(tab);
      }
    }
  });

  DDGsettings.ready().then(() => {
    httpsStorage.getLists(constants.httpsLists)
      .then(lists => https.setLists(lists))
      /* eslint-disable-next-line no-console */
      .catch(e => console.log(e));
  });
};

// Fire pixel on https upgrade failures to allow bad data to be removed from lists
chrome.webRequest.onErrorOccurred.addListener((e) => {
  if (!(e.type === 'main_frame')) {
    return;
  }

  const tab = tabManager.get({ tabId: e.tabId });

  // We're only looking at failed main_frame upgrades. A tab can send multiple
  // main_frame request errors so we will only look at the first one then set tab.hasHttpsError.
  if (!tab || !tab.mainFrameUpgraded || tab.hasHttpsError) {
    return;
  }

  if (e.error && e.url.match(/^https/)) {
    const errCode = constants.httpsErrorCodes[e.error];
    tab.hasHttpsError = true;

    if (errCode) {
      https.incrementUpgradeCount('failedUpgrades');
      // TODO - add our messaging?
      //            const url = new URL(e.url)
      //            pixel.fire('ehd', {
      //                'url': `${encodeURIComponent(url.hostname)}`,
      //                'error': errCode
      //            })
    }
  }
}, { urls: ['<all_urls>'] });

module.exports = {
  onStartup,
};
