// Inspired by the DuckDuckGo Privacy Extension (https://github.com/duckduckgo/duckduckgo-privacy-extension)
// Original source for this file can be found here:
// https://github.com/duckduckgo/duckduckgo-privacy-extension/tree/develop/shared/js/background/storage/

/*
 * Copyright (C) 2012, 2016 DuckDuckGo, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* For ESLint: List any global identifiers used in this file below */
/* global require log module chrome blockCounts exports Buffer */

'use strict';

const Dexie = require('dexie');
const load = require('./../load.es6');
const constants = require('../constants');
const DDGsettings = require('./../settings.es6');

class HTTPSStorage {
  constructor() {
    this.dbc = new Dexie(constants.httpsDBName);
    this.dbc.version(1).stores({
      httpsStorage: 'name,type,data,checksum',
    });
  }

  // Load https data defined in constants.httpsLists.
  // We wait until all promises resolve to send datd to https.
  // This is all or nothing. We gather data for each of the lists
  // and validate. If any list fails validation then promise.all will
  // reject the whole update.
  getLists() {
    return Promise.all(constants.httpsLists.map((list) => {
      const listCopy = JSON.parse(JSON.stringify(list));
      const etag = DDGsettings.getSetting(`${listCopy.name}-etag`) || '';

      return HTTPSStorage.getDataXHR(listCopy.url, etag).then((response) => {
        // for 200 response we update etags
        if (response && response.status === 200) {
          const newEtag = response.getResponseHeader('etag') || '';
          DDGsettings.updateSetting(`${listCopy.name}-etag`, newEtag);
        }

        // We try to process both 200 and 304 responses. 200s will validate
        // and update the db. 304s will try to grab the previous data from db
        // or throw an error if none exists.
        return this.processData(listCopy, response.data).then((resultData) => {
          if (resultData) {
            return resultData;
          }
          throw new Error(`HTTPS: process list xhr failed  ${listCopy.name}`);
        });
        /* eslint-disable-next-line no-unused-vars */
      }).catch(e => this.fallbackToDB(listCopy).then((backupFromDB) => {
        if (backupFromDB) {
          return backupFromDB;
        }
        // reset etag to force us to get fresh server data in case of an error
        DDGsettings.updateSetting(`${listCopy.name}-etag`, '');
        throw new Error(`HTTPS: data update for ${listCopy.name} failed`);
      }));
    }));
  }

  // validate xhr data and lookup previous data from local db if needed
  // verify the checksum before returning the processData result
  processData(listDetails, xhrData) {
    if (xhrData) {
      return HTTPSStorage.hasCorrectChecksum(xhrData).then((isValid) => {
        if (isValid) {
          this.storeInLocalDB(listDetails.name, listDetails.type, xhrData);
          return Object.assign(listDetails, xhrData);
        }
        return undefined;
      });
    }
    return Promise.resolve();
  }

  fallbackToDB(listDetails) {
    return this.getDataFromLocalDB(listDetails.name).then((storedData) => {
      if (!storedData) {
        return;
      }

      return HTTPSStorage.hasCorrectChecksum(storedData.data).then((isValid) => {
        if (isValid) {
          if (storedData && storedData.data) {
            return Object.assign(listDetails, storedData.data);
          }
        }
      });
    });
  }

  static getDataXHR(url, etag) {
    return load.loadExtensionFile({
      url, etag, returnType: 'json', source: 'external', timeout: 60000,
    });
  }

  getDataFromLocalDB(name) {
    log(`HTTPS: getting ${name} from db`);
    return this.dbc.open()
      .then(() => this.dbc.table('httpsStorage').get({ name }));
  }

  storeInLocalDB(name, type, data) {
    log(`HTTPS: storing ${name} in db`);
    return this.dbc.httpsStorage.put({ name, type, data });
  }

  static hasCorrectChecksum(data) {
    // not everything has a checksum
    if (!data.checksum) {
      return Promise.resolve(true);
    }

    // need a buffer to send to crypto.subtle
    const buffer = Buffer.from(data.data, 'base64');

    return crypto.subtle.digest('SHA-256', buffer).then((arrayBuffer) => {
      const sha256 = Buffer.from(arrayBuffer).toString('base64');
      if (data.checksum.sha256 && data.checksum.sha256 === sha256) {
        return true;
      }
      return false;
    });
  }
}
module.exports = new HTTPSStorage();
